import numpy as np
from sklearn.neighbors import NearestNeighbors

import torch
from torch_geometric.data import HeteroData

def calcu_adj(cord, cord2=None, neigh_cal ='knn', n_neigh = 8, n_radius=None, metric='minkowski'):
    '''
    Construct adjacency matrix with coordinates.
    input: cord, np.array
    '''
    
    if cord2 is None:
        cord2 = cord
        n_neigh += 1
    
    if neigh_cal == 'knn':
        neigh = NearestNeighbors(n_neighbors = n_neigh, metric = metric).fit(cord2)
        neigh_index = neigh.kneighbors(cord,return_distance=False)
        index = torch.LongTensor(np.vstack((np.repeat(range(cord.shape[0]),n_neigh),neigh_index.ravel())) ) 
    
    if neigh_cal == 'radius':
        neigh = NearestNeighbors(radius=n_radius, metric = metric).fit(cord2)
        dist, neigh_index = neigh.radius_neighbors(cord, return_distance=True)
        index = np.array([[],[]], dtype=int)
        
        for it in range(cord.shape[0]):
            index = np.hstack(((index, np.vstack((np.array([it]*neigh_index[it].shape[0]), neigh_index[it])))))  
        index = torch.LongTensor(index)
    
    value = torch.ones(index.shape[1])
    return torch.sparse.FloatTensor(index, value, torch.Size((cord.shape[0], cord2.shape[0])))


def hgat_data(adata, 
              batch_key, 
              batch_order = None,
              spatial_key ='spatial', 
              n_neigh_hom = 10, 
              n_radius_het = 0.1,
              kernal_thresh = None):
    
    if batch_order is None:
        batch_order = list(adata.obs[batch_key].value_counts().sort_index().index)

    # adjacency matrix within a slice
    feat_dict = {}
    adj_dict = {}
    coord_dict = {}
    index_dict = {}

    for batch_tmp in batch_order:
        adata_tmp = adata[adata.obs[batch_key]==batch_tmp].copy()
        feat_tmp = adata_tmp.obsm['latent']
        coord_tmp = adata_tmp.obsm[spatial_key]
        adj_tmp = calcu_adj(coord_tmp, 
                            neigh_cal = 'knn', 
                            n_neigh = n_neigh_hom, 
                            metric ='minkowski')._indices()
        feat_dict[batch_tmp] = feat_tmp
        coord_dict[batch_tmp] = coord_tmp
        adj_dict[batch_tmp, '0', batch_tmp] = adj_tmp
        index_dict[batch_tmp] = list(adata_tmp.obs_names)

    # adjacency matrix between slices
    cross_adj_dict = {}
    for target_tmp in batch_order:
        for source_tmp in batch_order:
            if target_tmp != source_tmp:
                cross_adj_dict[target_tmp, '1', source_tmp] = calcu_adj(feat_dict[target_tmp], 
                                                                        feat_dict[source_tmp], 
                                                                        neigh_cal ='radius', 
                                                                        n_radius = n_radius_het,
                                                                        metric ='cosine')._indices()

    adj_dict.update(cross_adj_dict)

    # construction of HeteroData
    data = HeteroData()
    for ii in list(feat_dict.keys()):
        data[ii].x = torch.from_numpy(feat_dict[ii]).float()

    for jj in list(adj_dict.keys()):
        data[jj].edge_index = adj_dict[jj]

    # construction of kernal matrix
    if kernal_thresh is not None:
        from scipy.spatial.distance import cdist
        kernals = dict()
        for node_key, node_coord in coord_dict.items():
            dist_tmp = np.sqrt(cdist(node_coord, node_coord, metric='euclidean'))
            dist_tmp = dist_tmp/np.percentile(dist_tmp, kernal_thresh, axis=0)
            kernal_tmp = np.exp(-dist_tmp)
            kernal_tmp = torch.from_numpy(kernal_tmp).float()    
            kernal_tmp = 0.5 * (kernal_tmp + kernal_tmp.T)
            kernals[node_key] = kernal_tmp
    
    else:
        kernals = None
    
    return data, kernals, index_dict



import numpy as np
import scanpy as sc
import os
import datetime
import torch


def set_seed(seed):
    import random 
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True


def construct_folder(path_name):
    result_path = path_name+'_'+datetime.datetime.now().strftime('%m-%d-%y %H:%M')
    if not os.path.exists(result_path):
        os.makedirs(result_path)
        os.makedirs(result_path + '/embedding')
        os.makedirs(result_path + '/embedding/train')
        os.makedirs(result_path + '/location')
        os.makedirs(result_path + '/location/edge')
    return result_path


class MakeLogClass:
    def __init__(self, log_file):
        self.log_file = log_file
        if os.path.exists(self.log_file):
            os.remove(self.log_file)
    def make(self, *args):
        # print(*args)
        # Write the message to the file
        with open(self.log_file, "a") as f:
            for arg in args:
                f.write("{}\r\n".format(arg))


def mclust_R(adata, num_cluster=10, modelNames='EEE', used_obsm='latent', random_seed=2022, key_add='clusters'):
    """
    Clustering using the mclust algorithm.
    The parameters are the same as those in the R package mclust.
    """
    np.random.seed(random_seed)
    import rpy2.robjects as robjects
    robjects.r.library("mclust")
    import rpy2.robjects.numpy2ri
    rpy2.robjects.numpy2ri.activate()
    r_random_seed = robjects.r['set.seed']
    r_random_seed(random_seed)
    rmclust = robjects.r['Mclust']
    res = rmclust(rpy2.robjects.numpy2ri.numpy2rpy(adata.obsm[used_obsm]), num_cluster, modelNames)
    mclust_res = np.array(res[-2])
    adata.obs[key_add] = mclust_res
    adata.obs[key_add] = adata.obs[key_add].astype('str')
    return adata



def cluster_func(adata, clustering, use_rep, res=1, cluster_num=None, key_add='cluster'):
    if clustering == 'louvain':
        sc.pp.neighbors(adata, use_rep=use_rep, key_added=key_add)
        sc.tl.louvain(adata, resolution=res, neighbors_key=key_add, key_added=key_add)
    if clustering == 'leiden':
        sc.pp.neighbors(adata, use_rep=use_rep, key_added=key_add)
        sc.tl.leiden(adata, resolution=res, neighbors_key=key_add, key_added=key_add)
    if clustering == 'kmeans':
        from sklearn.cluster import KMeans
        km = KMeans(n_clusters=cluster_num, random_state=2022).fit(adata.obsm[use_rep])
        adata.obs[key_add] = km.labels_
    if clustering == 'mclust':
        adata = mclust_R(adata, num_cluster=cluster_num, modelNames='EEE', used_obsm=use_rep, random_seed=2022, key_add=key_add)
    adata.obs[key_add] = adata.obs[key_add].astype('category')
    return adata





import torch
import torch.nn.functional as F
from torch import nn



# ============================================================================ #
# activation function

exp_act = lambda x: torch.exp(x)

def acti_fun(activate):
    if activate == 'relu':
        return F.relu
    elif activate == 'sigmoid':
        return torch.sigmoid
    elif activate == 'exp':
        return exp_act
    elif activate == 'softplus':
        return F.softplus
    elif activate =='tanh':
        return F.tanh


# ============================================================================ #
# layer

class FC_Layer(nn.Module):
    
    def __init__(self, in_features, out_features, bn=False, activate='relu', dropout=0.0):
        
        super(FC_Layer, self).__init__()
        
        self.in_features = in_features
        self.out_features = out_features
        
        self.bn = bn
        self.activate = activate
        self.dropout = dropout
        
        self.layer = nn.Linear(in_features, out_features)
        self.bn_layer = nn.BatchNorm1d(out_features)
    
    def forward(self, x):
        
        x = self.layer(x)
        if self.bn:
            x = self.bn_layer(x)
        
        if self.dropout!=0:
            return F.dropout(acti_fun(self.activate)(x), p=self.dropout, training=self.training)
        return acti_fun(self.activate)(x)       
    
    def __repr__(self):
        return self.__class__.__name__ + ' (' + str(self.in_features) + ' -> ' + str(self.out_features) + ')'


# ============================================================================ #
# NB & ZINB model


class New_NB_AE(nn.Module):
    
    def __init__(self, input_dim, hidden_dim, latent_dim, n_batch=None, dropout=0.2):
        
        super(New_NB_AE, self).__init__()
        
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.latent_dim = latent_dim
        
        self.dropout = dropout
        self.n_batch = n_batch
        
        self.input_dim_all = input_dim + n_batch if n_batch is not None else input_dim
        self.latent_dim_all = latent_dim + n_batch if n_batch is not None else latent_dim
        
        self.layer1 = FC_Layer(self.input_dim_all, self.hidden_dim, bn=True, activate='relu', dropout=self.dropout)
        self.layer2 = FC_Layer(self.hidden_dim, self.latent_dim, bn=True, activate='relu', dropout=self.dropout)
        self.layer3 = FC_Layer(self.latent_dim_all, self.hidden_dim, bn=False, activate='relu')
        self.layer4 = FC_Layer(self.hidden_dim, self.input_dim, bn=False, activate='exp')
        
        if self.n_batch is not None:
            self.layer_logi = torch.nn.Parameter(torch.randn(self.input_dim, self.n_batch))
        else:
            self.layer_logi = torch.nn.Parameter(torch.randn(self.input_dim))
    
    def forward(self, x, batch_tensor=None): 
        
        if self.n_batch is not None:
            assert batch_tensor.shape[1] == self.n_batch
            x = torch.cat((x, batch_tensor), dim=-1)
        z = self.layer2(self.layer1(x))
        
        if self.n_batch is not None:
            z = torch.cat((z, batch_tensor), dim=-1)  
        rate_scaled = self.layer4(self.layer3(z))
        
        if self.n_batch is not None:
            self.logi = F.linear(batch_tensor, self.layer_logi)
        else:
            self.logi = self.layer_logi
        
        return rate_scaled, self.logi.exp(), None, z
    


class New_ZINB_AE(nn.Module):
    
    def __init__(self, input_dim, hidden_dim, latent_dim, n_batch=None, dropout=0.2):
        
        super(New_ZINB_AE, self).__init__()
        
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.latent_dim = latent_dim
        
        self.dropout = dropout
        self.n_batch = n_batch
        
        self.input_dim_all = input_dim + n_batch if n_batch is not None else input_dim
        self.latent_dim_all = latent_dim + n_batch if n_batch is not None else latent_dim
        
        self.layer1 = FC_Layer(self.input_dim_all, self.hidden_dim, bn=True, activate='relu', dropout=self.dropout)
        self.layer2 = FC_Layer(self.hidden_dim, self.latent_dim, bn=True, activate='relu', dropout=self.dropout)
        self.layer3 = FC_Layer(self.latent_dim_all, self.hidden_dim, bn=False, activate='relu')
        self.layer_disp = FC_Layer(self.hidden_dim, self.input_dim, bn=False, activate='exp')
        self.layer_drop = FC_Layer(self.hidden_dim, self.input_dim, bn=False, activate='sigmoid')
        
        if self.n_batch is not None:
            self.layer_logi = torch.nn.Parameter(torch.randn(self.input_dim, self.n_batch))
        else:
            self.layer_logi = torch.nn.Parameter(torch.randn(self.input_dim))
    
    def forward(self, x, batch_tensor=None): 
        
        if self.n_batch is not None:
            assert batch_tensor.shape[1] == self.n_batch
            x = torch.cat((x, batch_tensor), dim=-1)
        z = self.layer2(self.layer1(x))
        
        if self.n_batch is not None:
            z = torch.cat((z, batch_tensor), dim=-1)       
        x3 = self.layer3(z)
        
        rate_scaled = self.layer_disp(x3)
        dropout = self.layer_drop(x3)
        if self.n_batch is not None:
            self.logi = F.linear(batch_tensor, self.layer_logi)
        else:
            self.logi = self.layer_logi
        
        return rate_scaled, self.logi.exp(), dropout, z
    

ae_dict = {'zinb':New_ZINB_AE, 'nb':New_NB_AE}





import torch
from torch import Tensor, nn
import torch.nn.functional as F

from torch_geometric.nn.conv import MessagePassing
from torch_geometric.nn.inits import glorot
from torch_geometric.typing import Adj, EdgeType, Metadata, NodeType, OptTensor
from torch_geometric.utils import softmax
from torch_geometric.nn import GATConv

from typing import Dict, List, Optional, Tuple, Union


# ============================================================================ #
# GAT Module

class GAT_pyg(nn.Module):
    
    def __init__(self, latent_dim=32, dropout_gat=0.5):
        super(GAT_pyg, self).__init__()
        
        self.dropout = dropout_gat
        self.gat1 = GATConv(latent_dim, latent_dim, 1)
        self.gat2 = GATConv(latent_dim, latent_dim, 1)
    
    def forward(self, x, edge_index, get_attention=False):
        
        x = self.gat1(x, edge_index)
        x = F.elu(x)
        
        z = F.dropout(x, self.dropout, training=self.training)
        
        if get_attention:
            xbar, atte_tmp = self.gat2(z, edge_index, return_attention_weights=True)
        else:  
            xbar = self.gat2(z, edge_index)
        
        xbar = F.elu(xbar)
        xbar = F.dropout(xbar, self.dropout, training=self.training)
        
        if get_attention:
            return xbar, z, atte_tmp
        return xbar, z


# ============================================================================ #
# corss graph attention

# def group(
#     x_gat_,
#     xs: Dict,
#     k_lin: nn.Module,
# ) -> Tuple[OptTensor, OptTensor]:
#     if len(xs) == 0:
#         return None, None    
#     else:
#         num_edge_types = len(xs)
#         out = torch.stack(list(xs.values()))     
#         if out.numel() == 0:
#             return out.view(0, out.size(-1)), None
#         attn_score = (torch.tanh(k_lin(x_gat_.mean(1)).mean(0)) * torch.tanh(k_lin(out)).mean(1)).sum(-1)
#         attn = F.softmax(attn_score, dim=0)
#         out = torch.sum(attn.view(num_edge_types, 1, -1) * out, dim=0)
#     return out, attn


def group(
    xs: Dict,
    q: nn.Parameter,
    k_lin: nn.Module,
) -> Tuple[OptTensor, OptTensor]:
    if len(xs) == 0:
        return None, None
    else:
        num_edge_types = len(xs)
        out = torch.stack(list(xs.values()))
        if out.numel() == 0:
            return out.view(0, out.size(-1)), None
        attn_score = (q * torch.tanh(k_lin(out)).mean(1)).sum(-1)
        attn = F.softmax(attn_score, dim=0)
        out = torch.sum(attn.view(num_edge_types, 1, -1) * out, dim=0)
        return out, attn


# def group(
#     xs: List[Tensor],
#     q: nn.Parameter,
#     k_lin: nn.Module,
# ) -> Tuple[OptTensor, OptTensor]:
#     if len(xs) == 0:
#         return None, None
#     else:
#         num_edge_types = len(xs)
#         out = torch.stack(xs)
#         if out.numel() == 0:
#             return out.view(0, out.size(-1)), None
#         attn_score = (q * torch.tanh(k_lin(out)).mean(1)).sum(-1)
#         attn = F.softmax(attn_score, dim=0)
#         out = torch.sum(attn.view(num_edge_types, 1, -1) * out, dim=0)
#         return out, attn


class HGAT(MessagePassing):
    
    def __init__(
        self,
        num_channels: Union[int, Dict[str, int]],
        metadata: Metadata,
        heads: int = 1,
        negative_slope=0.2,
        dropout_hom: float = 0.7,
        dropout_het: float = 0.5,
        gamma: float = 0.9,
        **kwargs,
    ):
        super().__init__(aggr='add', node_dim=0, **kwargs)
        
        self.heads = heads
        self.num_channels = num_channels
        self.negative_slope = negative_slope
        self.metadata = metadata
        self.dropout_hom = dropout_hom
        self.dropout_het = dropout_het
        self.gamma = gamma
        self.k_lin = nn.Linear(num_channels, num_channels)
        self.q = nn.Parameter(torch.Tensor(1, num_channels))
        self.lin_src = nn.ParameterDict()
        self.lin_dst = nn.ParameterDict()  
        for edge_type in metadata[1]:
            if edge_type[0] != edge_type[-1]:            
                edge_type = '__'.join(edge_type)
                self.lin_src[edge_type] = nn.Parameter(torch.Tensor(1, heads, num_channels))
                self.lin_dst[edge_type] = nn.Parameter(torch.Tensor(1, heads, num_channels))
        # print(self.lin_src)
        # print(self.lin_dst)
        self.gat_layer = GAT_pyg(latent_dim=num_channels, dropout_gat=self.dropout_hom)
        self.reset_parameters()
    
    def reset_parameters(self):
        glorot(self.lin_src)
        glorot(self.lin_dst)
        self.k_lin.reset_parameters()
        glorot(self.q)
    
    def forward(
        self,
        x_dict: Dict[NodeType, Tensor],
        edge_index_dict: Dict[EdgeType, Adj],
        return_semantic_attention_weights: bool = False,
        get_attention: bool = False
    ):
        
        H, D = self.heads, self.num_channels
        x_gat_dict, x_hat_dict = {}, {}      # intra-slice embedding; inter-slice embedding 
        cell_attn_dict = {}

        # Intra-slices aggregation
        for node_type, x in x_dict.items():
            if get_attention:
                xbar_tmp, z_tmp, atte_tmp = self.gat_layer( x, 
                                                            edge_index_dict[node_type, '0', node_type],
                                                            get_attention = True)
                cell_attn_dict[(node_type, '0', node_type)] = atte_tmp

            else:
                xbar_tmp, z_tmp = self.gat_layer( x, edge_index_dict[node_type, '0', node_type])
            
            x_gat_dict[node_type] = xbar_tmp.view(-1, H, D)
            x_hat_dict[node_type] = {}    
        
        # Inter-slices aggregation
        for edge_type, edge_index in edge_index_dict.items():
            
            src_type, _, dst_type = edge_type
            
            if src_type != dst_type:
                
                edge_type = '__'.join(edge_type)
                
                lin_src = self.lin_src[edge_type]
                lin_dst = self.lin_dst[edge_type]
                
                x_src = x_gat_dict[src_type]
                x_dst = x_gat_dict[dst_type]
                
                alpha_src = (x_src * lin_src).sum(dim=-1)
                alpha_dst = (x_dst * lin_dst).sum(dim=-1)
                
                # propagate_type: (x_dst: PairTensor, alpha: PairTensor)
                out = self.propagate(edge_index, 
                                    x=(x_src, x_dst),
                                    alpha=(alpha_src, alpha_dst), size=None)

                x_hat_dict[dst_type][src_type] = F.relu(out)

                if get_attention:
                    # N1, N2 = x_src.shape[0], x_dst.shape[0]
                    # alpha = alpha_src.repeat(N2, 1).T + alpha_dst.repeat(N1, 1)
                    # alpha = F.leaky_relu(alpha, self.negative_slope)
                    # alpha = softmax(alpha, index, ptr, size_i)
                    cell_attn_dict[(dst_type, '1', src_type)] = (x_src, x_dst)

        
        # aggregating from other slices
        semantic_attn_dict = {}
        
        for node_type, x_hat_ in x_hat_dict.items():
            # x_gat_ = x_gat_dict[node_type]
            x_hat_out, attn = group(x_hat_, self.q, self.k_lin)
            # x_hat_out, attn = group(x_gat_, x_hat_, self.k_lin)
            x_hat_dict[node_type] = x_hat_out
            semantic_attn_dict[node_type] = attn
        
        out_gat_dict = {key:value.sum(1) for key, value in x_gat_dict.items()}
        out = {key:self.gamma*out_gat_dict[key] + (1-self.gamma)*x_hat_dict[key] for key in x_hat_dict.keys()}
        
        if get_attention:
            if return_semantic_attention_weights:
                return out, semantic_attn_dict, cell_attn_dict
            return out, cell_attn_dict
        
        else:
            if return_semantic_attention_weights:
                return out, semantic_attn_dict
            return out
    
    def message(self, x_j: Tensor, alpha_i: Tensor, alpha_j: Tensor,
                index: Tensor, ptr: Optional[Tensor],
                size_i: Optional[int]) -> Tensor:
        
        alpha = alpha_j + alpha_i
        alpha = F.leaky_relu(alpha, self.negative_slope)
        alpha = softmax(alpha, index, ptr, size_i)
        alpha = F.dropout(alpha, p=self.dropout_het, training=self.training)
        out = x_j * alpha.view(-1, self.heads, 1)
        
        return out.view(-1, self.num_channels)
    
    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}({self.num_channels}, '
                f'heads={self.heads})')







import os
import numpy as np
import pandas as pd
from tqdm import tqdm
import matplotlib.pyplot as plt

import torch
from torch.utils.data import DataLoader
import torch.nn.functional as F


from STAIR.embedding.module_ae import ae_dict
from STAIR.embedding.module_hgat import HGAT
from STAIR.embedding.dataset_ae import MyDataset
from STAIR.embedding.dataset_hgat import hgat_data
from STAIR.embedding.loss import nll_loss

from STAIR.utils import MakeLogClass


class Emb_Align(object):

    """
    
    Spatial embedding alignment of multiplt ST slices, including CE alignment and SE alignment. 
    They perform embedding alignment using AE module and HGAT module, respectively.

    Parameters
    ----------
    adata
        AnnData object object of scanpy package
    batch_key
        The key containing slice information in .obs
    hvg
        Slice order used to perform alignment. Align according to the default order of elements in batch_key if None
    n_hidden
        The number of hidden dimension in the AE module
    n_latent
        The number of latent dimension in the AE module
    dropout_rate
        Dropout rate
    likelihood
        Distribution assumptions for expression matrices, 'nb' or 'zinb'
    
    Examples
    --------
    >>> adata = sc.read_h5ad(path_to_anndata)
    >>> emb_align = Emb_Align(adata, batch_key='batch', result_path=result_path)
    >>> emb_align.prepare()
    >>> emb_align.preprocess()
    >>> emb_align.latent()
    >>> emb_align.prepare_hgat(batch_order=keys_use)
    >>> emb_align.train_hgat()
    >>> adata, atte = emb_align.predict_hgat()

    """
    
    def __init__(
        self,
        adata,
        batch_key = None,
        hvg = False,
        n_hidden: int = 128,
        n_latent: int = 32,
        dropout_rate: float = 0.2,
        likelihood: str = "nb",
        device: str = None, 
        num_workers: int = 4,
        result_path = None, 
        make_log: bool = True
    ):
        super(Emb_Align, self).__init__() 

        if device is not None:
            self.device = device
        else:
            self.device = 'cuda:0' if torch.cuda.is_available() else 'cpu'
        
        self.adata = adata
        self.n_batch = None if batch_key is None else len(set(adata.obs[batch_key]))
        self.batch_key = batch_key
        
        if hvg:
            self.n_input = hvg
        else:
            self.n_input = adata.shape[1]
        
        self.hvg = hvg
        self.n_latent = n_latent
        self.likelihood = likelihood
        
        self.ae = ae_dict[self.likelihood]( input_dim = self.n_input, 
                                            hidden_dim = n_hidden, 
                                            latent_dim = n_latent, 
                                            dropout = dropout_rate,
                                            n_batch = self.n_batch 
                                            ).to(self.device)        
        
        self.num_workers = num_workers
        self.result_path = result_path
        self.make_log = make_log
        
        if self.make_log:
            self.makeLog = MakeLogClass(f"{self.result_path}/log_emb.tsv").make
            self.makeLog(f"Module parameter set of AE")
            self.makeLog(f"  Likelihood: {self.likelihood}")
            self.makeLog(f"  Input dim: {self.n_input}")
            self.makeLog(f"  Latent Dim: {self.n_latent}")
            self.makeLog(f"  Dropout: {dropout_rate}")


    def prepare(self, count_key=None, lib_size='explog', normalize=True, scale=False):
       
        '''
        Prepare data for analysis.
        
        Parameters
        ----------
        count_key
            Layer to normalize instead of X. If None, X is normalized.
        lib_size
            The way to calculate the library size for each cell.
        normalize
            Whether to normalize the expression, default is True.
        scale
            Whether to scale the expression, default is False.
        
        '''
        self.data_ae = MyDataset(self.adata, 
                                 count_key = count_key,
                                 size = lib_size, 
                                 normalize = normalize, 
                                 hvg = self.hvg, 
                                 scale = scale, 
                                 batch_key = self.batch_key)
        
        if self.hvg:
            self.adata = self.adata[:,:self.hvg]
        
        if self.make_log:
            self.makeLog(f"  Library size: {lib_size}")
            self.makeLog(f"  Input normalize: {normalize}")
            self.makeLog(f"  Input scale: {scale}")
            self.makeLog(f"  Hvg: {self.hvg}")

    def preprocess( self,
                    lr = 0.001, 
                    weight_decay = 0, 
                    epoch_ae = 100, 
                    batch_size = 64,
                    plot = False):
        '''
        Preprocessed training part using AE module
        
        Parameters
        ----------
        weight_decay
            Weight_decay of training, default is 0.
        epoch_ae
            Total epoch of training, default is 40.
        batch_size
            The number of samples used for each gradient update in mini-batch training, default is 128.
        plot
            Whether to plot the loss in each epoch, default is False.
        '''
        if self.make_log:
            self.makeLog(f"Training parameter set of AE")
            self.makeLog(f"  Learning rate: {lr}")
            self.makeLog(f"  Weight decay: {weight_decay}")
            self.makeLog(f"  Training epoch: {epoch_ae}")
            self.makeLog(f"  Batch size: {batch_size}")
            
        data_loader = DataLoader(self.data_ae, shuffle=True, batch_size=batch_size, num_workers=self.num_workers)
        optimizer = torch.optim.Adam(self.ae.parameters(), lr=lr, weight_decay=weight_decay)   

        train_loss = []
        for epoch in tqdm(range(0, epoch_ae)):
            loss_tmp = 0
            if self.batch_key is not None:
                for i, (feat_tmp, count_tmp, size_tmp, batch_tmp) in enumerate(data_loader):
                    feat_tmp = feat_tmp.to(self.device)
                    count_tmp = count_tmp.to(self.device)
                    size_tmp = size_tmp.to(self.device)
                    batch_tmp = batch_tmp.to(self.device)
                    self.ae.train()
                    rate_scaled_tmp, logits_tmp, drop_tmp, _ = self.ae(feat_tmp, batch_tmp)
                    rate_tmp = rate_scaled_tmp*size_tmp
                    mean_tmp = rate_tmp*logits_tmp
                    optimizer.zero_grad()
                    loss_train = nll_loss(count_tmp, mean_tmp, rate_tmp, drop_tmp, dist=self.likelihood).mean()
                    loss_train.backward()
                    optimizer.step()
                    # if i%5 == 0:
                        # print("AE Epoch:{},  loss {}".format(epoch,loss_train.item()))
                    loss_tmp += loss_train.item()
                train_loss.append(loss_tmp/len(data_loader))
            else:
                for i, (feat_tmp, count_tmp, size_tmp) in enumerate(data_loader):
                    feat_tmp = feat_tmp.to(self.device)
                    count_tmp = count_tmp.to(self.device)
                    size_tmp = size_tmp.to(self.device)
                    self.ae.train()
                    rate_scaled_tmp, logits_tmp, drop_tmp, _ = self.ae(feat_tmp)
                    rate_tmp = rate_scaled_tmp*size_tmp
                    mean_tmp = rate_tmp*logits_tmp
                    optimizer.zero_grad()
                    loss_train = nll_loss(count_tmp, mean_tmp, rate_tmp, drop_tmp, dist=self.likelihood).mean()
                    loss_train.backward()
                    optimizer.step()
                    # if i%5 == 0:
                        # print("AE Epoch:{},  loss {}".format(epoch,loss_train.item()))
                    loss_tmp += loss_train.item()
                train_loss.append(loss_tmp/len(data_loader))

        self.ae.eval()

        # torch.save(self.ae.state_dict(),f'{self.result_path}/embedding/train/model_pre.pth')

        # if plot:
        #     plt.plot(train_loss)
        #     plt.savefig(f'{self.result_path}/embedding/train/loss_pre.png')
        #     plt.close()

    def latent( self,
                batch_size = None,
                return_data = False
              ):
        '''
        Preprocessed predicting part using AE module
        
        Parameters
        ----------
        batch_size
            Batch size in predicting part.
        return_data
            Whether to return adata, default is False.
        '''
        self.ae.eval()
        
        if batch_size is None:
            batch_size = len(self.data_ae)
        
        dataloader = DataLoader(self.data_ae, shuffle=False, batch_size=batch_size, num_workers=self.num_workers)

        z = torch.empty(size=[0,self.n_latent])
        mean = torch.empty(size=[0,self.n_input])

        with torch.no_grad():
            
            if self.batch_key is not None:                
                for _, (feat_tmp, count_tmp, lib_tmp, batch_tmp) in enumerate(dataloader):
                    # print(i)
                    feat_tmp = feat_tmp.to(self.device)
                    count_tmp = count_tmp.to(self.device)
                    lib_tmp = lib_tmp.to(self.device)
                    batch_tmp = batch_tmp.to(self.device)
                    rate_scaled_tmp, logits_tmp, _, z_tmp = self.ae(feat_tmp, batch_tmp)
                    z = torch.cat([z, z_tmp.cpu()[:,:self.n_latent]])
                    rate_tmp = rate_scaled_tmp*lib_tmp
                    mean_tmp = rate_tmp*logits_tmp
                    mean = torch.cat([mean, mean_tmp.cpu()])
            else:                
                for _, (feat_tmp, count_tmp, lib_tmp) in enumerate(dataloader):
                    # print(i)
                    feat_tmp = feat_tmp.to(self.device)
                    count_tmp = count_tmp.to(self.device)
                    lib_tmp = lib_tmp.to(self.device)
                    rate_scaled_tmp, logits_tmp, _, z_tmp = self.ae(feat_tmp)
                    z = torch.cat([z, z_tmp.cpu()[:,:self.n_latent]])
                    rate_tmp = rate_scaled_tmp*lib_tmp
                    mean_tmp = rate_tmp*logits_tmp
                    mean = torch.cat([mean, mean_tmp.cpu()])

        self.adata.obsm['latent'] = z.detach().cpu().numpy()
        self.adata.layers['Denoise'] = mean.detach().cpu().numpy()

        if return_data:
            return self.adata
    
    def prepare_hgat(self,
                     slice_key = None,
                     slice_order = None,
                     spatial_key = 'spatial',
                     n_neigh_hom = 10, 
                     c_neigh_het = 0.9,
                     kernal_thresh = 0.):
        '''
        Construct heterogeneous graph for HAT module.

        Parameters
        ----------
        slice_key
            Key of slice information in .obs.
        slice_order
            List with slice names ordered by the physical location.
        spatial_key
            Key of raw spatial location of spots in .obsm.
        n_neigh_hom
            Number of neighbors based on location in the same slice, default is 10.
        c_neigh_het
            Similarity cutoff based on expression latent in the defferent slice, default is 0.9.
        '''
        if slice_key is None:
            slice_key = self.batch_key
        
        self.data_hgat, self.kernals, self.index_dict = hgat_data(  self.adata, 
                                                                    batch_key = slice_key, 
                                                                    batch_order = slice_order, 
                                                                    spatial_key = spatial_key, 
                                                                    n_neigh_hom = n_neigh_hom, 
                                                                    n_radius_het = 1-c_neigh_het,
                                                                    kernal_thresh = kernal_thresh)
        
        if self.make_log:
            self.makeLog(f"Module parameter set of HGAT")
            self.makeLog(f"  Spatial key: {spatial_key}")
            self.makeLog(f"  Neighbor number of intra-slice: {n_neigh_hom}")
            self.makeLog(f"  Similarity cutoff of inter-slice: {c_neigh_het}")
            
    def train_hgat( self,
                    gamma: float = 0.8,
                    epoch_hgat: int = 150,
                    re_weight: float = 1.,
                    si_weight: float = 0.,
                    lr: float = 0.001, 
                    weight_decay: float = 0.,
                    negative_slope: float = 0.2,
                    dropout_hom: float = 0.5,
                    dropout_het: float = 0.5,
                    plot = False):
        '''
        Training step HAT module.

        Parameters
        ----------
        gamma
            Weight of homogeneous representation in SE. U=λ∙U^hom+(1-λ)∙U^het
        epoch_hgat
            Total epoch of training, default is 100.
        dropout_hom
            Dropout rate in aggregating intra-slice information.
        dropout_het
            Dropout rate in aggregating inter-slice information.
        plot
            Whether to plot the loss in each epoch, default is False.
        
        '''
        if self.make_log:
            self.makeLog(f"  Dropout_hom: {dropout_hom}")
            self.makeLog(f"  Dropout_het: {dropout_het}")
            self.makeLog(f"Training parameter set of HGAT")
            self.makeLog(f"  Weight of intra-slice: {gamma}")
            self.makeLog(f"  Reconstruction weight: {re_weight}")
            self.makeLog(f"  Similarity weight: {si_weight}")
            self.makeLog(f"  Training epoch: {epoch_hgat}")
            self.makeLog(f"  Learning rate: {lr}")
            self.makeLog(f"  Weight decay: {weight_decay}")
            self.makeLog(f"  Negative slope: {negative_slope}")
 
        self.hgat = HGAT( num_channels = self.n_latent, 
                          metadata = self.data_hgat.metadata(), 
                          negative_slope = negative_slope,
                          dropout_hom = dropout_hom,
                          dropout_het = dropout_het,
                          gamma = gamma
                         ).to(self.device)  

        self.data_hgat = self.data_hgat.to(self.device)  
        optimizer_hgat = torch.optim.Adam(self.hgat.parameters(), lr=lr, weight_decay=weight_decay)

        loss_list = []
        for epoch in tqdm(range(epoch_hgat)):
            
            self.hgat.train()
            
            out = self.hgat(self.data_hgat.x_dict, 
                            self.data_hgat.edge_index_dict, 
                            return_semantic_attention_weights=False)
            
            re_loss = 0
            si_loss = 0
            
            for node_type, x in out.items():
                re_loss += F.mse_loss(x, self.data_hgat[node_type].x)
                if si_weight != 0:
                    si_loss += F.mse_loss(self.kernals[node_type].to(self.device), torch.mm(x,x.T))
            
            if si_weight != 0:
                loss = re_weight * re_loss + si_weight * si_loss  
                loss_list.append([loss.item(), re_loss.item(), si_loss.item()])
            else:
                loss = re_weight * re_loss
                loss_list.append([loss.item(), re_loss.item()])
            
            # print("HGAT Epoch:{}  loss:{} ".format(epoch, loss.item()))
            
            loss.backward()
            optimizer_hgat.step()
            optimizer_hgat.zero_grad()

        self.hgat.eval()
        
        # if not os.path.exists(self.result_path + '/embedding/train'):
        #     os.makedirs(self.result_path + '/embedding/train')
            
        # torch.save(self.hgat.state_dict(),f'{self.result_path}/embedding/train/model_hgat.pth')

        if plot:
            if si_weight != 0:
                ls = plt.plot(loss_list)
                plt.legend(handles=ls, labels=['loss', 're_loss', 'si_loss'], loc='best')
                plt.savefig(f'{self.result_path}/embedding/train/loss_hgat.png')
                plt.close()   
            else:
                ls = plt.plot(np.array(loss_list)[:,:2].tolist())
                plt.legend(handles=ls, labels=['loss', 're_loss'], loc='best')
                plt.savefig(f'{self.result_path}/embedding/train/loss_hgat.png')
                plt.close()   
           

    def predict_hgat(self, get_attention=False):
        '''
        Predicting step HAT module.

        '''
        with torch.no_grad():
    
            self.hgat.eval()
            if get_attention:
                out, atte, atte_node = self.hgat(self.data_hgat.x_dict,
                                                self.data_hgat.edge_index_dict, 
                                                return_semantic_attention_weights=True,
                                                get_attention=True)
            else:
                out, atte = self.hgat(self.data_hgat.x_dict,
                                        self.data_hgat.edge_index_dict, 
                                        return_semantic_attention_weights=True)
                    
        # HAN_SE
        out = [pd.DataFrame(values.detach().cpu().numpy(), 
                            index=self.index_dict[ii]) for ii, values in out.items()]
        
        out = pd.concat(out)
        self.adata.obsm['STAIR'] = out.loc[self.adata.obs_names].values

        # Attention between slices
        # i = 0
        # atte_ = []
        # for _, value_ in atte.items():
        #     value_ = value_.cpu().detach()
        #     value_ = value_.unsqueeze(0)
        #     i += 1
        #     atte_.append(value_)
        
        i = 0
        atte_ = []
        for _, value_ in atte.items():
            value_ = value_.cpu().detach()
            value_ = torch.concat([value_[:i], torch.tensor([1.0]), value_[i:]])
            value_ = value_.unsqueeze(0)
            i += 1
            atte_.append(value_)

        atte_ = torch.concat(atte_)
        atte_ = pd.DataFrame(atte_, index=atte.keys(), columns=atte.keys())
    
        if get_attention:
            return self.adata, atte_, atte_node
        return self.adata, atte_
    
    
    


from setuptools import setup, find_packages

setup(
    name='STAIR-tools',  
    packages=find_packages(),
    include_package_data=True,
    package_data={
        'STAIR': ['ABAanno/ccfv3/*.nrrd', 
                  'ABAanno/*.csv', 
                  'ABAanno/*.pkl'],
    },
    version='1.1.13',
    description='Spatial Transcriptomic Alignment, Integration, and de novo 3D Reconstruction',
    author='yuanyuanyu',
    license='MIT',
    URL='https://github.com/yuyuanyuana/STAIR'
)